from abc import ABC


class BaseStreamState(ABC):
    """Filler class in case we want to add abstract methods in the future,
    such as maybe requiring a method to clear/reset/copy StreamStates..."""
    pass



"""Code which changes the way Python processes imports. Allows for capsules to
import Python modules and packages within the capsule itself.
"""
from importlib.abc import Loader, MetaPathFinder
from importlib.machinery import ModuleSpec
import os
from pathlib import Path
from zipfile import ZipFile


class ZipFinder(MetaPathFinder):
    """An import finder that allows modules and packages inside a zip file to be
    imported.
    """

    def __init__(self, zip_file: ZipFile, capsule_dir_path: Path):
        """
        :param zip_file: The ZipFile loaded in memory
        :param capsule_dir_path:
            The path to the directory where the development version of the
            capsule is stored. For example, if the capsule being loaded is in
            ../capsules/my_capsule_name.py

            Then capsule_dir_path would be
            ../capsules/my_capsule_name/

            The reason for this is so that debugging can work within capsules.
            When a capsule is executed, it is "compiled" to this path, so that
            debug information still works (along with breakpoints!)
        """
        self._zip_file = zip_file
        self._capsule_dir_path = capsule_dir_path
        self._capsule_dir_name = self._capsule_dir_path.name

    def find_spec(self, fullname, _path=None, _target=None):
        if not self._in_capsule(fullname):
            # If the capsule directory name is not in the fullname, it's
            # not our job to import it
            return None

        if fullname == self._capsule_dir_name:
            # If the capsule root directory is being loaded, return a
            # modulespec
            return ModuleSpec(
                name=fullname,
                loader=None,
                is_package=True)

        # Get rid of the capsule name prefix
        pruned_fullname = _remove_capsule_name(self._capsule_dir_name,
                                               fullname)
        package_path = _package_fullname_to_path(pruned_fullname)
        module_path = _module_fullname_to_path(pruned_fullname)

        if package_path in self._zip_file.namelist():
            # If a directory exists with a name that matches the import, we
            # assume it is a package import.
            return ModuleSpec(name=fullname,
                              loader=None,
                              is_package=True)
        elif module_path in self._zip_file.namelist():
            # If a .py file exists with a name that matches the import, we
            # assume it is a module import
            module_file_path = self._capsule_dir_path / module_path
            loader = ZipModuleLoader(zip_file=self._zip_file,
                                     module_file_path=module_file_path,
                                     capsule_dir_name=self._capsule_dir_name)
            return ModuleSpec(name=fullname,
                              loader=loader)

        raise ImportError(f"Problem while importing {fullname}")

    def _in_capsule(self, fullname):
        parts = fullname.split(".")
        if parts[0] != self._capsule_dir_name:
            return False
        return True


class ZipModuleLoader(Loader):
    """Loads modules from a zip file."""

    def __init__(self, zip_file: ZipFile,
                 module_file_path: Path,
                 capsule_dir_name: str):
        """
        :param zip_file: The ZipFile loaded in memory
        :param module_file_path: The path to where the python file would be
            in the filesystem if it was being run directly as opposed to in a
            *.cap zip.
        """
        self._zip_file = zip_file
        self._module_file_path = module_file_path
        self._capsule_dir_name = capsule_dir_name

    def create_module(self, spec):
        return None

    def exec_module(self, module):
        pruned_name = _remove_capsule_name(
            capsule_name=self._capsule_dir_name,
            fullname=module.__name__)
        zip_path = _module_fullname_to_path(pruned_name)

        # Extract the code from the zip
        code = self._zip_file.read(zip_path)

        # Compile code with the file path set first, so that debugging and
        # tracebacks work in development (they reference a file)
        # also breakpoints work in IDE's, which is quite helpful.
        compiled = compile(code, self._module_file_path, "exec")
        exec(compiled, module.__dict__)
        return module


def _package_fullname_to_path(fullname):
    """Converts a package's fullname to a file path that should be the package's
    directory.

    :param fullname: The fullname of a package, like package_a.package_b
    :return: A derived filepath, like package_a/package_b
    """
    return fullname.replace(".", os.sep) + os.sep


def _module_fullname_to_path(fullname):
    """Converts a module's fullname to a file path that should be the module's
    Python file.

    :param fullname: The fullname of a module, like package_a.my_module
    :return: A derived filepath, like package_a/my_module.py
    """
    return fullname.replace(".", os.sep) + ".py"


def _remove_capsule_name(capsule_name, fullname):
    """Remove "capsule_name" from capsule_name.some_module.some_module2
    Since the files in the zip file won't have "capsule_name" in the paths
    """
    parts = fullname.split(".")
    return ".".join(parts[1:])



from enum import Enum
from typing import List, Dict, Union

from .detection_node import DetectionNode

DETECTION_NODE_TYPE = Union[None, DetectionNode, List[DetectionNode]]


class NodeDescription:
    """Describes required aspects of a DetectionNode. This is used to specify
    what kind of DetectionNodes a capsule takes as input and what kind of
    DetectionNodes the capsule outputs.

    For example, a capsule that encodes people and face detections would use
    NodeDescriptions to define its inputs and outputs like so:

    >>> input_type = NodeDescription(
    ...     detections=["person", "face"])
    >>> output_type = NodeDescription(
    ...     detections=["person", "face"],
    ...     encoded=True)

    A capsule that uses a car's encoding to classify the color of a car would
    look like this.

    >>> input_type = NodeDescription(
    ...     detections=["car"],
    ...     encoded=True)
    >>> output_type = NodeDescription(
    ...     detections=["car"],
    ...     attributes={"color": ["blue", "yellow", "green"]},
    ...     encoded=True)

    A capsule that detects dogs and takes no existing input would look like
    this.

    >>> input_type = NodeDescription(size=NodeDescription.Size.SINGLE)
    >>> output_type = NodeDescription(
    >>>                 size=NodeDescription.Size.ALL,
    >>>                 detections=["dog"])
    """

    class Size(Enum):
        NONE = 1
        """A NodeDescription that does not take in or produce DetectionNodes.
        """
        SINGLE = 2
        """A NodeDescription that takes in or produces a single DetectionNode.
        """
        ALL = 3
        """A NodeDescription that takes in all available DetectionNodes or
        produces any number of DetectionNodes.
        """

    def __init__(self, *,
                 size: Size,
                 detections: List[str] = None,
                 attributes: Dict[str, List[str]] = None,
                 encoded: bool = False,
                 tracked: bool = False,
                 extra_data: List[str] = None):
        """
        :param size: The number of DetectionNodes that this capsule either
            takes as input or provides as output
        :param detections: A list of acceptable detection class names. A node
            that meets this description must have a class name that is present
            in this list
        :param attributes: A dict whose key is the classification type and
            whose value is a list of possible attributes. A node that meets
            this description must have a classification for each classification
            type
        :param encoded: If true, the DetectionNode must be encoded to meet this
            description
        :param tracked: If true, the DetectionNode is being tracked
        :param extra_data: A list of keys in a NodeDescription's extra_data. A
            DetectionNode that meets this description must have extra data for
            each name listed here.
        """
        self.size = size
        self.detections = detections if detections else []
        self.attributes = attributes if attributes else {}
        self.encoded = encoded
        self.tracked = tracked
        self.extra_data = extra_data if extra_data else []

    def difference(self, other: "NodeDescription"):
        """Outputs a NodeDescription that represents the change between this
        NodeDescription and the given one. For instance, if the given
        NodeDescription has been detected for people and this NodeDescription
        has not, the resulting NodeDescription would look like this.

        >>> NodeDescription(size=NodeDescription.Size.ALL,
        >>>                 detections=["person"])

        This can be used to compare a capsule's input and output nodes to see
        what operation the capsule provides.
        """
        # Find the classification difference between this NodeDescription and\
        # the other
        # Found here: https://stackoverflow.com/a/32815681
        new_attributes = {
            k: other.attributes[k]
            for k in set(other.attributes) - set(self.attributes)}

        # Find the difference between this NodeDescription and the other
        new_detections = list(set(other.detections) - set(self.detections))

        # Find the difference between the extra data
        new_extra_data = list(set(other.extra_data) - set(self.extra_data))

        # Set encoded in the diff description to true only if this
        # NodeDescription was not encoded and the other one was
        newly_encoded = not self.encoded and other.encoded

        # Set tracked in diff description to true if this node was not tracked
        # and the other one was
        newly_tracked = not self.tracked and other.tracked

        return NodeDescription(
            size=other.size,
            detections=new_detections,
            attributes=new_attributes,
            encoded=newly_encoded,
            tracked=newly_tracked,
            extra_data=new_extra_data)

    def describes(self, node: DetectionNode):
        """Check if a DetectionNode has the information described in this
        NodeDescription. """

        if len(self.detections) > 0 and node.class_name not in self.detections:
            return False

        if self.encoded and node.encoding is None:
            return False

        if self.tracked and node.track_id is None:
            return False

        if len(self.extra_data):
            all_data_available = all(key in node.extra_data
                                     for key in self.extra_data)
            if not all_data_available:
                return False

        if self.attributes != {}:
            if node.attributes == {}:
                return False

            for category, possible_values in self.attributes.items():
                if category not in node.attributes.keys():
                    return False

                if node.attributes[category] not in possible_values:
                    good_values = self.attributes[category]
                    bad_value = node.attributes[category]
                    raise ValueError(
                        f"Invalid attribute value in DetectionNode! "
                        f"DetectionNode has value '{bad_value}' for attribute "
                        f"category '{category}', but the NodeDescription "
                        f"indicates that only values {good_values} are "
                        f"supported. Consider adding '{bad_value}' to the "
                        f"list of possible values for attribute '{category}'.")

        return True

    def __repr__(self):
        rep = self.__dict__.copy()
        if not self.encoded:
            rep.pop("encoded")
        if not self.extra_data:
            rep.pop("extra_data")
        if not self.detections:
            rep.pop("detections")
        if not self.attributes:
            rep.pop("attributes")
        if not self.tracked:
            rep.pop("tracked")
        return str(rep)

    def __eq__(self, other):
        return self.__dict__ == other.__dict__



"""Defines ovens.

Ovens are classes that know how to do some kind of generic work in batches.
"""
import logging
import queue
from queue import Queue
from threading import Thread
from typing import Any, Callable, List, NamedTuple, Optional


class _OvenRequest(NamedTuple):
    """A request that is sent to an oven to do some work on an image, and
    push predictions into the output_queue

    output_queue: The queue for the oven to put the results in
    img_bgr: An OpenCV BGR image to run detection on
    """
    output_queue: Queue
    input_data: Any


class Oven:
    """This class simplifies receiving work from a multitude of sources and
    running that work in a batched predict function, then returning that
    work to the respective output queues."""

    def __init__(self, batch_fn: Callable[[List[Any]], List[Any]],
                 max_batch_size=40,
                 num_workers: int = 1):
        """Initialize a new oven.
         that the oven will wait between running a batch regardless of batch
         size.

        :param batch_fn: A function that takes in a list of inputs and iterates
        the outputs in the same order as the inputs.
        :param max_batch_size: The maximum length of list to feed to batch_fn
        :param num_workers: How many workers should be calling batch_fn
        """
        self.batch_fn = batch_fn
        self.max_batch_size = max_batch_size
        self._request_queue = Queue()
        self.workers = [Thread(target=self._worker,
                               daemon=True,
                               name="OvenThread")
                        for _ in range(num_workers)]

        # The number of images currently in the work queue or being processed
        self._num_imgs_being_processed: int = 0

        self._running: bool = True

        for worker in self.workers:
            worker.start()

    @property
    def total_imgs_in_pipeline(self) -> int:
        return self._request_queue.qsize() + self._num_imgs_being_processed

    def submit(self, input_data: Any, output_queue: Queue = None) -> Queue:
        """Creates an OvenRequest for you and returns the output queue"""
        output_queue = output_queue if output_queue else Queue()
        self._request_queue.put(_OvenRequest(
            output_queue=output_queue,
            input_data=input_data))
        return output_queue

    def _on_requests_ready(self, batch: List[_OvenRequest]) -> None:
        """Push images through the given prediction backend

        :param batch: A list of requests to work on
        """
        # Extract the images from the requests
        inputs: List[Any] = [req.input_data for req in batch]
        output_queues: List[Queue] = [req.output_queue for req in batch]

        # Route the results to each request
        predictions: List[Any] = self.batch_fn(inputs)

        response_count = 0
        for response_count, output in enumerate(predictions, start=1):
            output_queues[response_count - 1].put(output)

        if response_count != len(inputs):
            logging.error(f"CRITICAL ERROR: Backend returned {response_count} "
                          f"responses. Expected {len(inputs)}")

    def _worker(self):
        self._running = True

        while self._running:
            # Get a new batch
            batch = self._get_next_batch()

            # If no batch was able to be retrieved, restart the loop
            if batch is None:
                continue

            # Check to make sure the thread isn't trying to end
            if not self._running:
                break

            self._num_imgs_being_processed += len(batch)
            self._on_requests_ready(batch)
            self._num_imgs_being_processed -= len(batch)

        self._running = False

    def _get_next_batch(self) -> Optional[List[_OvenRequest]]:
        """A helper function to help make the main thread loop more readable
        :returns: A non-empty list of collected items, or None if the worker is
                  no longer running (i.e. self._continue == False)
        """
        batch: List[_OvenRequest] = []
        while len(batch) < self.max_batch_size:
            # Check if there's a new request
            try:
                # Try to get a new request. Have a timeout to check if closing
                new_request = self._request_queue.get(timeout=.1)
            except queue.Empty:
                # If the thread is being requested to close, exit early
                if not self._running:
                    return None

                # Wait for requests again
                continue

            batch.append(new_request)

            # If the request queue is now empty, let worker run everything in
            # the batch
            if self._request_queue.empty():
                break

        return batch

    def close(self) -> None:
        """Stop the oven gracefully."""
        self._running = False
        for worker in self.workers:
            worker.join()



from .capsule import BaseCapsule
from .stream_state import BaseStreamState
from .backend import BaseBackend
from .detection_node import DetectionNode, rect_to_coords, BoundingBox
from .node_description import NodeDescription, DETECTION_NODE_TYPE
from .device_mapping import DeviceMapper
from .loading.capsule_loading import load_capsule
from .modifiers import Crop, Clamp, Resize, SizeFilter
from .options import (
    FloatOption,
    EnumOption,
    IntOption,
    BoolOption,
    Option,
    common_detector_options,
    OPTION_TYPE
)
from .loading.packaging import (
    CAPSULE_EXTENSION,
    package_capsule
)
from .loading.errors import (
    InvalidCapsuleError,
    IncompatibleCapsuleError
)


import logging
from threading import RLock
from typing import Callable, List

import tensorflow as tf
from tensorflow.python.client import device_lib

_devices = None
_devices_lock = RLock()


def get_all_devices() -> List[str]:
    """Returns a list of devices in the computer. Example:
    ['CPU:0', 'XLA_GPU:0', 'XLA_CPU:0', 'GPU:0', 'GPU:1', 'GPU:2', 'GPU:3']
    """
    global _devices, _devices_lock

    # Initialize the device list if necessary
    with _devices_lock:
        if _devices is None:
            # We set these config options on the session because otherwise list
            # local_devices allocates all of the available GPU memory to the
            # computer
            #
            # Note that this config has the side effect of being set for all
            # future sessions#
            #
            # TODO: Use tf.config.list_physical_devices in TF 2.1
            # TODO: Remove the config when using TF_FORCE_GPU_ALLOW_GROWTH\
            #  environment variable
            _process_gpu_options = tf.GPUOptions(allow_growth=True)
            config = tf.ConfigProto(gpu_options=_process_gpu_options)

            with tf.Session(config=config):
                all_devices = device_lib.list_local_devices()

            # Get the device names and remove duplicates, just in case...
            tf_discovered_devices = {
                d.name.replace("/device:", "")
                for d in all_devices
            }

            # Discover devices using OpenVINO
            try:
                from openvino.inference_engine import IECore
                ie = IECore()
                openvino_discovered_devices = {
                    d for d in ie.available_devices
                    if not d.lower().startswith("cpu")
                }
            except ModuleNotFoundError:
                logging.warning("OpenVINO library not found. "
                                "OpenVINO devices will not be discovered. ")
                openvino_discovered_devices = set()
            _devices = list(tf_discovered_devices
                            | openvino_discovered_devices)
    return _devices


class DeviceMapper:
    def __init__(self, filter_func: Callable[[List[str]], List[str]]):
        """The filter will take in a list of devices formatted as
        ["CPU:0", "CPU:1", "GPU:0", "GPU:1"], etc and output a filtered list of
        devices.
        """
        self.filter_func = filter_func

    @staticmethod
    def map_to_all_gpus(cpu_fallback=True) -> 'DeviceMapper':
        def filter_func(devices):
            gpu_devices = [d for d in devices if d.startswith("GPU:")]
            if not gpu_devices and cpu_fallback:
                return ["CPU:0"]
            return gpu_devices

        return DeviceMapper(filter_func=filter_func)

    @staticmethod
    def map_to_single_cpu() -> 'DeviceMapper':
        def filter_func(devices):
            return [next(d for d in devices if d.startswith("CPU:"))]

        return DeviceMapper(filter_func=filter_func)

    @staticmethod
    def map_to_all_myriad(cpu_fallback=True):
        """Returns Myriad device. If cpu_fallback=True, it will return CPU:0
        if there are no myriad devices."""

        def filter_func(devices):
            myriad_devices = [d for d in devices
                              if d.lower().startswith("myriad")]
            if not myriad_devices and cpu_fallback:
                return ["CPU:0"]
            return myriad_devices

        return DeviceMapper(filter_func=filter_func)



import threading


def verify_all_threads_closed(allowable_threads=None):
    """A convenient function to throw an error if all threads are not closed"""
    if allowable_threads is None:
        allowable_threads = []
    allowable_threads += ['pydevd.Writer',
                          'pydevd.Reader',
                          'pydevd.CommandThread',
                          'profiler.Reader',
                          'MainThread']

    open_threads = [t.name for t in threading.enumerate()
                    if t.name not in allowable_threads]

    if len(open_threads) != 0:
        raise EnvironmentError(
            "Not all threads were shut down! Currently running threads: " +
            str(open_threads))



from pathlib import Path

import mock

from vcap import BaseCapsule
from vcap.loading.capsule_loading import load_capsule


def load_capsule_with_one_device(packaged_capsule_path: Path) -> BaseCapsule:
    """
    Load the capsule, but patch out the DeviceMapper so that it never returns
    multiple devices.

    Essentially, this disables capsules from loading backends onto multiple
    devices, for example GPUs. This is for the purpose of speeding up test
    setup and teardown.
    """

    def mock_init(self, filter_func):
        # Don't call the superclass init so that self.filter_func isn't
        # overridden
        self.filter_func = lambda all_devices: [filter_func(all_devices)[0]]

    with mock.patch('vcap.device_mapping.DeviceMapper.__init__',
                    mock_init):
        capsule: BaseCapsule = load_capsule(path=packaged_capsule_path)

    return capsule



from Cryptodome.Cipher import AES


def encrypt(password, source, dest):
    """Encrypt the given source to dest.

    :param password: The password to encrypt with
    :param source: A file-like object to read from
    :param dest: A file-like object to write results to
    """
    cipher = AES.new(password.encode('utf-8'), AES.MODE_EAX)
    ciphertext, tag = cipher.encrypt_and_digest(source.read())

    [dest.write(x) for x in (cipher.nonce, tag, ciphertext)]


def decrypt(password, source):
    """Decrypt the given source into bytes.

    :param password: The password to decrypt with
    :param source: A file-like object to read from
    :return: The resulting bytes
    """
    nonce, tag, ciphertext = [source.read(x) for x in (16, 16, -1)]

    cipher = AES.new(password.encode("utf-8"), AES.MODE_EAX, nonce)
    return cipher.decrypt_and_verify(ciphertext, tag)


def decrypt_file(relative_path, password):
    """Decrypt the file given by the path.

    :param relative_path: A Path object describing the file to decrypt
    :param password: The password to decrypt with
    """
    path = relative_path
    with open(path, "rb") as f:
        return decrypt(password, f)


def encrypt_file(in_path, out_path, password):
    """Encrypt the file as a new file.

    Note that in_path should NOT be the same file as out_path, because the files
    are never fully loaded into memory.

    :param in_path: A Path object, the location of the unencrypted file
    :param out_path: A Path object, the location to save the encrypted data
    """
    with open(in_path, "rb") as in_file, open(out_path, "wb") as out_file:
        encrypt(password, in_file, out_file)



def cache(method):
    """Caches the result of a method call. Caches are specific to the instance
    of the object that this method is for.
    """

    def on_call(self, *args, **kwargs):
        name = method.__name__
        try:
            return self._cache[name]
        except AttributeError:
            # Create the cache if necessary
            self._cache = {}
        except KeyError:
            # Handled below
            pass

        val = method(self, *args, **kwargs)
        self._cache[name] = val
        return val

    return on_call



import logging
import random
from abc import ABC, abstractmethod
from threading import RLock
from typing import Dict, List, Optional

import numpy as np

from .backend import BaseBackend
from .caching import cache
from .device_mapping import DeviceMapper, get_all_devices
from .node_description import DETECTION_NODE_TYPE, NodeDescription
from .options import Option, OPTION_TYPE
from .stream_state import BaseStreamState


class BaseCapsule(ABC):
    """An abstract base class that all capsules must subclass. Defines the
    interface that capsules are expected to implement.

    A capsule is a zipped and encrypted collection of files and configuration
    information. It represents a single algorithm with a specific set
    of functionality, like an inception_resnet_v2 behavior classification
    model.

    A class that subclasses from this class is expected to be defined in a
    capsule.py file in a capsule.
    """

    stream_state = BaseStreamState
    """This attribute is the basic 'stream_state object that is initialized
    for every new stream that a capsule is run on, and de-initialized when that
    stream goes away. It is intended to be overridden by capsules that have 
    stateful operations across a single stream."""

    backends: Optional[List[BaseBackend]] = None
    """A list of the backends the capsule has initialized"""

    backends_lock = RLock()
    """Lock the list of backends whenever it's being accessed"""

    def __init__(self, capsule_files: Dict[str, bytes], inference_mode=True):
        """Load a capsule file.

        :param capsule_files: A dict of {"file_name": FILE_BYTES} of the files
            that were found and loaded in the capsule
        :param inference_mode: If True, the model will be loaded and the
            backends will start for it. If False, the capsule will never be
            able to run inference, but it will still have it's various readable
            attributes.
        """

        if inference_mode:
            # We use type(self) to avoid having capsules need to put
            # staticmethod on each of their backend loaders.
            # self.backend_loader is a function not a method because we usually
            # use the following system for creating subclasses:
            #
            # >>> class CoolCapsule(BaseCapsule):
            # ...     ...
            # ...     backend_loader = lambda: some_func(...)
            #
            # This prevents self from being passed as the first (and wrong)
            # argument. We may want to change this in the future.
            load_backend = lambda device: type(self).backend_loader(
                capsule_files=capsule_files,
                device=device)

            # Initialize the backends on the capsules requested devices
            # for example: GPU:0, GPU:1, CPU:0, MYRIAD, HDDL
            all_devices = get_all_devices()
            devices_to_load_to = self.device_mapper.filter_func(all_devices)

            # Throw an error if the capsule couldn't load
            if not devices_to_load_to:
                message = f"The capsule was not able to find any valid " \
                          f"devices to initialize on! The devices " \
                          f"discovered on this machine are: {all_devices}."
                raise EnvironmentError(message)

            logging.info(f"Loading capsule {self.name} "
                         f"onto devices {', '.join(devices_to_load_to)}")
            self.backends = list(map(load_backend, devices_to_load_to))

        # Keep a dictionary for keeping track of different StreamState objs
        self._stream_states: Dict[int, BaseStreamState] = {}
        """{stream_id: BaseStreamState}"""

        self._stream_state_lock = RLock()

    def __repr__(self):
        rep = f"Capsule(name={self.name}, " \
              f"input={self.input_type}, " \
              f"output={self.output_type})"
        return rep

    def __del__(self):
        self.close()

    def process_frame(self,
                      frame: np.ndarray,
                      detection_node: DETECTION_NODE_TYPE,
                      options: Dict[str, OPTION_TYPE],
                      state: BaseStreamState) \
            -> DETECTION_NODE_TYPE:
        """Find the Backend that has an oven with the least amount of images
        in the pipeline, and run with that Backend. In multi-GPU scenarios,
        this is the logic that allows even usage across all GPUs."""
        with self.backends_lock:
            # BaseBackend.oven.total_imgs_in_pipeline only represents the
            # images in a single capsule's pipeline, not all the capsules.
            # This helps distribute the load of multiple capsules.
            #
            # Note the inplace shuffle, but this should be okay, esp. w/ lock
            random.shuffle(self.backends)
            laziest_backend: BaseBackend \
                = min(self.backends,
                      key=lambda backend: backend.oven.total_imgs_in_pipeline)

        return laziest_backend.process_frame(
            frame, detection_node, options, state)

    def get_state(self, stream_id: int) -> BaseStreamState:
        """Get or create the StreamState for this stream_id and return it."""
        with self._stream_state_lock:
            # Adds stream state if stream_id is not already a key
            return self._stream_states.setdefault(stream_id,
                                                  self.stream_state())

    def clean_up(self, stream_id: int) -> None:
        """Dereference the StreamState that relates to this stream_id"""
        with self._stream_state_lock:
            if stream_id in self._stream_states:
                del self._stream_states[stream_id]

    def close(self) -> None:
        """This method MUST close the oven first, then de-initialize the
        backend so as to clear up memory.
        """
        if self.backends is not None:
            with self.backends_lock:
                for backend in self.backends:
                    backend.close()
                    backend.oven.close()
                self.backends = None

    @property
    @cache
    def capability(self) -> NodeDescription:
        """Returns a NodeDescription showing the difference between the input
        and output NodeDescriptions for this capsule. This tells the caller
        what service this capsule provides. For instance, a capsule that
        classifies a person's gait would return the following:

        >>> NodeDescription(
        ...     size=NodeDescription.Size.ALL,
        ...     attributes={"Gait": ["walking", "running"]})

        The fact that the capsule's input is a person is not shown in this
        value because person detection is not something this capsule provides.
        """
        return self.input_type.difference(self.output_type)

    @property
    @cache
    def default_options(self):
        """Return a dict of key->value where the value is the default value
        for this capsules options"""
        return {key: val.default for key, val in self.options.items()}

    # These things are to be filled in by a subclassing capsule
    @property
    def name(self) -> str:
        """A name to uniquely identify the capsule."""
        raise NotImplementedError

    @property
    def version(self) -> int:
        """The version of the capsule.

        When should you bump the version of a capsule? When:
            - You've changed the usage of existing capsule options
            - You've changed the model or algorithm
            - You've changed the input/output node descriptions

        When shouldn't you bump the version of a capsule? When:
            - You only did code restructuring
            - You've updated the capsule to work with a newer API version
            - You've added (but not removed or changed previous) capsule
              options

        In summary, the version is most useful for differentiating a capsule
        from its previous versions.
        """
        raise NotImplementedError

    @property
    def description(self) -> str:
        """A human-readable description of what the capsule does."""
        return ""

    @property
    def device_mapper(self) -> DeviceMapper:
        return DeviceMapper.map_to_all_gpus()

    @staticmethod
    @abstractmethod
    def backend_loader(capsule_files: Dict[str, bytes], device: str) \
            -> "BaseBackend":
        """Initializes the capsule and returns the initialized Backend.

        :param capsule_files: A dict of
            filename: file_bytes of the files within the capsule's root
                      directory.
        :param device: A device, in the format of "GPU:0", "CPU:2"
        The devices passed into the capsule is decided by the
        Capsule.device_mapper
        """
        raise NotImplementedError

    @property
    def input_type(self) -> NodeDescription:
        """Describes the type of node this capsule expects to work with."""
        raise NotImplementedError

    @property
    def output_type(self) -> NodeDescription:
        """Describes the type of node this capsule will output."""
        raise NotImplementedError

    @property
    def options(self) -> Dict[str, Option]:
        """A list of zero or more options that can be configured at runtime."""
        return {}



from typing import List, Dict, Optional, Union
from uuid import UUID

import numpy as np

from .caching import cache


class BoundingBox:
    def __init__(self, x1, y1, x2, y2):
        self.x1 = x1
        self.y1 = y1
        self.x2 = x2
        self.y2 = y2

    @property
    def size(self):
        """
        :return: The width and height of the bounding box
        """
        return self.x2 - self.x1, self.y2 - self.y1

    @property
    def rect(self):
        """Get the bounding box in 'rect' format, as (x1, y1, x2, y2)"""
        return self.x1, self.y1, self.x2, self.y2

    @property
    def xywh(self):
        """Return the bounding box in 'xywh' format, as
        (x1, y1, width, height)"""
        return (self.x1, self.y1,
                self.x2 - self.x1,
                self.y2 - self.y1)

    @property
    def center(self):
        return (self.x1 + (self.x2 - self.x1) / 2,
                self.y1 + (self.y2 - self.y1) / 2)

    @property
    def height(self):
        return self.y2 - self.y1

    @property
    def width(self):
        return self.x2 - self.x1

    def __eq__(self, other):
        return other.__dict__ == self.__dict__


class DetectionNode:
    """A bounding box detection and any corresponding information about that
    detection. DetectionNodes may form a tree with children and parent nodes to
    express that a detection is "part" of another detection.
    """

    def __init__(self, *, name: str,
                 coords: List[List[Union[int, float]]],
                 attributes: Dict[str, str] = None,
                 children: List['DetectionNode'] = None,
                 encoding: Optional[np.ndarray] = None,
                 track_id: Optional[UUID] = None,
                 extra_data: Dict[str, object] = None):
        """
        :param name: The class name, describing what the detection is of, like
            "person"
        :param coords: A list of coordinates defining a shape
        :param attributes: A dictionary of {"category": "value"} pairs.
            ie, {"Gender": "Male", "Age": "Young}
        :param children: Child DetectionNodes that are a "part" of the parent,
            for instance, a head DetectionNode might be a child of a person
            DetectionNode
        :param encoding: The encoding for this detection, if any
        :param track_id: If this object is tracked, this is the unique
        identifier for this detection node that ties it to other detection nodes
        in future and past frames (within the same stream).
        :param extra_data: Any other domain-specific data that describe
            this detection. These values are not used for computation.
        """
        self.class_name = name
        self.coords = coords
        self.attributes = attributes if attributes else {}
        self.encoding = encoding
        self.track_id = track_id
        self.children = children if children else []
        self.extra_data = extra_data if extra_data else {}

        self._bbox = None

    def scale(self, scale_amount_x: float, scale_amount_y: float):
        """Scales the detection coordinates of the tree by the given scales.

        :param scale_amount_x: The amount to scale x by
        :param scale_amount_y: The amount to scale y by
        """
        for i, c in enumerate(self.coords):
            self.coords[i] = [round(c[0] * scale_amount_x),
                              round(c[1] * scale_amount_y)]
        self._bbox = None

        if self.children is not None:
            for child in self.children:
                child.scale(scale_amount_x, scale_amount_y)

    def offset(self, offset_x: int, offset_y: int):
        """Offsets the detection coordinates of the tree by the given offsets.

        :param offset_x: The amount to offset x by
        :param offset_y: The amount to offset y by
        """
        for i, c in enumerate(self.coords):
            self.coords[i] = [c[0] - offset_x, c[1] - offset_y]
        self._bbox = None

        if self.children is not None:
            for child in self.children:
                child.offset(offset_x, offset_y)

    @property
    @cache
    def all_attributes(self):
        """Return all attributes including child attributes"""

        def get_attrs(child):
            attributes = child.attributes
            for child in child.children:
                attributes.update(get_attrs(child))
            return attributes

        return get_attrs(self)

    def __repr__(self):
        rep = {
            "class_name": self.class_name,
            "track_id": self.track_id,
            "attributes": self.attributes,
            "extra_data": self.extra_data,
            "coords": [(round(x), round(y)) for x, y in self.coords],
            "encoding": "Encoded" if self.encoding is not None else None
        }

        return str(rep)

    @property
    def bbox(self) -> BoundingBox:
        if self._bbox is None:
            self._bbox = self._make_bbox()
        return self._bbox

    def _make_bbox(self):
        """
        :return: Create a fully containing bounding box of the detection
            polygon
        """
        sorted_x = sorted([c[0] for c in self.coords])
        sorted_y = sorted([c[1] for c in self.coords])
        return BoundingBox(sorted_x[0], sorted_y[0],
                           sorted_x[-1], sorted_y[-1])


def rect_to_coords(rect):
    """Converts a rect in the [x1, y1, x2, y2] format to coordinates."""
    return [[rect[0], rect[1]],
            [rect[2], rect[1]],
            [rect[2], rect[3]],
            [rect[0], rect[3]]]



import gc
import logging
import random
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from typing import Dict, List, Union, Optional
from uuid import uuid4

import cv2
import mock
import numpy as np

from vcap import (
    NodeDescription,
    DetectionNode,
    BaseCapsule,
    OPTION_TYPE,
    FloatOption,
    IntOption,
    BoolOption,
    EnumOption)
from vcap.loading.packaging import CAPSULE_EXTENSION, package_capsule
from vcap.testing import load_capsule_with_one_device
from vcap.testing.thread_validation import verify_all_threads_closed

NUM_STREAMS = 5
"""Number of random batches of images to send to the capsule"""

NUM_BATCH_CASES = 5
"""Number of batches of fuzzed inputs to perform on different images"""


def make_detection_node(frame_shape,
                        node_description: NodeDescription) -> DetectionNode:
    """Creates a fake detection node that describes the given node description.
    :param frame_shape: The shape of the frame in (height, width, channels)
    :param node_description: The description that the returned node must
        adhere to
    :return: A fake detection node that adheres to this description
    """
    height, width, _ = frame_shape
    attributes = {category: random.choice(possible_values)
                  for category, possible_values in
                  node_description.attributes.items()}
    extra_data = {data_key: 0.5129319283
                  for data_key in node_description.extra_data}

    # Create random coordinates for this detection
    x1 = random.randint(0, width - 3)
    y1 = random.randint(0, height - 3)
    x2 = x1 + random.randint(0, width - x1 + 1) + 2
    y2 = y1 + random.randint(0, height - y1 + 1) + 2

    return DetectionNode(
        name=random.choice(node_description.detections),
        coords=[[x1, y1], [x2, y1], [x2, y2], [x1, y2]],
        attributes=attributes,
        encoding=np.zeros((128,)) if node_description.encoded else None,
        track_id=uuid4() if node_description.tracked else None,
        extra_data=extra_data)


def make_capsule_options(capsule: BaseCapsule) -> Dict[str, OPTION_TYPE]:
    """Create a random set of options that are valid within the specs that the
    capsule options allow"""
    options = {}
    for name, opt in capsule.options.items():
        possible_vals = [opt.default]
        if isinstance(opt, (FloatOption, IntOption)):
            min_max_vals = [opt.min_val, opt.max_val]
            # If min/max are unrestricted (None) remove them
            possible_vals += [val for val in min_max_vals if val is not None]

        elif isinstance(opt, BoolOption):
            possible_vals += [True, False]
        elif isinstance(opt, EnumOption):
            possible_vals += opt.choices
        else:
            raise ValueError(
                f"This test needs to be updated for the new capsule option "
                f"type: {opt}")
        options[name] = random.choice(list(set(possible_vals)))
    return options


def _run_inference_on_images(images: List[np.ndarray], capsule: BaseCapsule):
    """Run inference on a list of images"""
    # Make multiple parallel requests with different images and different
    # input DetectionNodes to the capsule.
    request_input = []

    with ThreadPoolExecutor(
            thread_name_prefix="InputOutputValidation ") \
            as executor:
        for stream_id, image in enumerate(images):
            if capsule.input_type.size == NodeDescription.Size.NONE:
                input_node = None

            elif capsule.input_type.size == NodeDescription.Size.SINGLE:
                input_node = make_detection_node(image.shape,
                                                 capsule.input_type)

            elif capsule.input_type.size == NodeDescription.Size.ALL:
                input_node = [make_detection_node(image.shape,
                                                  capsule.input_type)
                              for _ in range(random.randint(0, 5))]
            else:
                raise NotImplementedError(
                    "The capsule did not have a NodeDescription.Size that was "
                    "known!")

            options = make_capsule_options(capsule)

            future = executor.submit(
                capsule.process_frame,
                frame=image,
                detection_node=input_node,
                options=options,
                state=capsule.stream_state())

            request_input.append((future, input_node))

        for future, input_node in request_input:
            # Postprocess the results
            prediction = future.result(timeout=90)

            # Verify that the capsule performed correctly

            if isinstance(prediction, DetectionNode):
                # Validate that what was returned by the capsule was valid
                assert (capsule.output_type.size
                        is NodeDescription.Size.SINGLE)
                output_nodes = [prediction]
            elif prediction is None and isinstance(input_node, DetectionNode):
                # If the capsule didn't output something, then it must have
                # modified the input node in-place. Validate the changes.
                assert (capsule.output_type.size
                        is NodeDescription.Size.SINGLE)
                output_nodes = [input_node]
            elif prediction is None and isinstance(input_node, list):
                # If this capsule accepts size ALL as input, then it must
                # have modified the detections within the input list
                assert capsule.input_type.size is NodeDescription.Size.ALL
                assert capsule.output_type.size is NodeDescription.Size.ALL
                output_nodes = input_node
            elif isinstance(prediction, list):
                # Validate that every detection node in the list is correct
                assert capsule.output_type.size is NodeDescription.Size.ALL
                output_nodes = prediction
            else:
                raise RuntimeError(f"Unknown prediction type: {prediction}")

            # Validate every output node against the capsules output_type
            for output_node in output_nodes:
                assert capsule.output_type.describes(output_node), \
                    ("Capsule failed to output a prediction that matches "
                     "the NodeDescription it had for it's output type. "
                     f"Prediction: {prediction}")
                # If this capsule can encode things, verify that the backend
                # correctly implemented the "distance" function
                if (capsule.capability.encoded
                        and prediction is not None
                        and len(prediction) > 0):
                    # Get one of the predictions
                    pred = prediction[0]

                    # Measure the distance from an encoding to itself
                    # (should be 0)
                    distances = capsule.backends[0].distances(
                        pred.encoding, np.array([pred.encoding]))
                    assert len(distances) == 1
                    assert distances[0] == 0, \
                        ("This assertion can be removed in the case that "
                         "there is some new distance function where two "
                         "encodings that are equal no longer have a distance "
                         "of 0. Until that case exists, keep this assertion.")


def _test_capsule_input_output(capsule, image_paths):
    """Creates necessary resources to load a capsule, then runs the
    necessary batches to test the capsule. """

    # Since we use fuzzing for some tests, we might as well set a seed
    random.seed(capsule.name)

    # Tests multiple cases for multiple images as input
    loaded_images = [cv2.imread(str(path)) for path in image_paths]
    try:
        # Run the test cases
        for _ in range(NUM_BATCH_CASES):
            images = random.choices(loaded_images, k=NUM_STREAMS)
            random.shuffle(images)
            _run_inference_on_images(images, capsule)
    except Exception as e:
        # If an error happens, clean up the threads. Otherwise, we will let
        # the rest of the test (garbage collection) make sure the capsule gets
        # closed properly.
        capsule.close()
        raise e


def perform_capsule_tests(unpackaged_capsule_dir: Union[Path, str],
                          image_paths: List[Union[Path, str]],
                          allowable_threads: Optional[List[str]] = None):
    """This tests several many things:
    1) The capsule can handle batches of images
    2) The capsule can handle different capsule option combinations
    3) The capsule handles the input type that it specifies
    4) The capsule does not keep references to itself, thus causing it not to
       be garbage collected
    """
    unpackaged_capsule_dir = Path(unpackaged_capsule_dir)

    logging.info("Testing capsule name: ", unpackaged_capsule_dir.name)
    packaged_capsule_path = (unpackaged_capsule_dir
                             .with_name(unpackaged_capsule_dir.stem)
                             .with_suffix(CAPSULE_EXTENSION))
    package_capsule(unpackaged_capsule_dir, packaged_capsule_path)

    with mock.patch.object(BaseCapsule, '__del__') as patched_del:
        capsule = load_capsule_with_one_device(packaged_capsule_path)
        _test_capsule_input_output(capsule, image_paths)

        # Since the __del__ method was wrapped, mock will not actually call the
        # underlying method when the object gets garbage collected- thus the
        # capsule will never get 'close' called on it, thus the oven threads
        # will not close. To fix that, we call 'close' now, and check that
        # __del__ WOULD have been called.
        capsule.close()

        referrers = gc.get_referrers(capsule)
        assert len(referrers) <= 1, \
            "No one else should have a reference to this capsule anymore! " \
            f"There were {len(referrers)} references from: {referrers}. " \
            f"Capsule in question: {capsule}"
        assert patched_del.call_count == 0

        # Get rid of the last reference to the capsule
        del capsule

        gc.collect()

        assert patched_del.call_count == 1, \
            "Oh no! This capsule is probably keeping a reference to itself. " \
            "This is a big no no!"

    verify_all_threads_closed(allowable_threads)



from abc import ABC, abstractmethod
from typing import List, Optional, Union


OPTION_TYPE = Union[int, float, bool, str]


def _check_number_range(value, min_val, max_val):
    error = ValueError(f"Value {value} is not in "
                       f"{min_val} <= x <= {max_val}")

    if min_val is not None and value < min_val:
        raise error
    if max_val is not None and value > max_val:
        raise error


class Option(ABC):
    """A class that all options subclass. Added mostly for documentation an
    type hinting purposes.
    """

    def __init__(self, default, description):
        self.default = default
        self.description: Optional[str] = description
        self.check(default)

    @abstractmethod
    def check(self, value):
        pass


class FloatOption(Option):
    """A capsule option that holds a floating point value with defined
    boundaries.

    Can have min_val or max_val be None"""

    def __init__(self, *, default: float,
                 min_val: Optional[float],
                 max_val: Optional[float],
                 description: Optional[str] = None):
        self.min_val = min_val
        self.max_val = max_val

        super().__init__(default=default, description=description)

    def check(self, value):
        """Checks the given value to see if it fits this option. Raises a
        ValueError in case of issues.
        :param value: The value to check
        """
        if not isinstance(value, float):
            raise ValueError(f"Expecting type float, got {type(value)}")
        _check_number_range(value, self.min_val, self.max_val)


class IntOption(Option):
    """A capsule option that holds an integer value.

    Can have min_val and max_val be None
    """

    def __init__(self, *, default: int,
                 min_val: Optional[int],
                 max_val: Optional[int],
                 description: Optional[str] = None):
        self.min_val = min_val
        self.max_val = max_val

        super().__init__(default=default, description=description)

    def check(self, value):
        """Checks the given value to see if it fits this option. Raises a
        ValueError in case of issues.
        :param value: The value to check
        """
        if not isinstance(value, int):
            raise ValueError(f"Expecting type int, got {type(value)}")
        _check_number_range(value, self.min_val, self.max_val)


class EnumOption(Option):
    """A capsule option that holds a choice from a discrete set of string
    values.
    """

    def __init__(self, *, default: str, choices: List[str],
                 description: Optional[str] = None):
        assert len(choices) > 0

        self.choices = choices

        super().__init__(default=default, description=description)

    def check(self, value):
        """Checks the given value to see if it fits this option. Raises a
        ValueError in case of issues.
        :param value: The value to check
        """
        if not isinstance(value, str):
            raise ValueError(f"Expecting type float, got {type(value)}")

        if value not in self.choices:
            raise ValueError(f"Value {value} is not one of {self.choices}")


class BoolOption(Option):
    """A capsule option that holds an boolean value."""

    def __init__(self, *, default: bool,
                 description: Optional[str] = None):
        super().__init__(default=default, description=description)

    def check(self, value):
        if not isinstance(value, bool):
            raise ValueError(f"Expecting type bool, got {type(value)}")


def check_option_values(capsule, option_vals):
    """Checks the given option values against the capsule's available options to
    make sure they are of the right name, type, and fit within the constraints.
    """
    for name, val in option_vals.items():
        if name not in capsule.options:
            raise ValueError(f"'{name}' is not a valid option for this capsule")
        option = capsule.options[name]
        try:
            option.check(val)
        except ValueError as e:
            raise ValueError(f"For option {name}: {e}")


"""Options that detector capsules tend to have
to avoid breaking API too often, make these options be short, with minimal 
keys. Try to add a name that is descriptive to the use case of the capsule.

If an existing one is changed, it is the responsibility of the person
changing to verify that all capsules have been updated. This is considered
a breaking capsule API change if the key is changed.
"""

common_detector_options = {
    "threshold": FloatOption(
        default=0.5,
        min_val=0.0,
        max_val=1.0,
        description="The confidence threshold for the detector to return a "
                    "detection for an object. Lower => more detections, "
                    "higher means fewer but more accurate detections."),
}



import configparser
import logging
import re
import sys
from io import BytesIO
from pathlib import Path
from types import ModuleType
from typing import Callable, List, Optional, Union
from zipfile import ZipFile

from vcap import BaseCapsule, BaseBackend, BaseStreamState, NodeDescription
from vcap.loading.errors import IncompatibleCapsuleError, InvalidCapsuleError

from .crypto_utils import decrypt_file
from .import_hacks import ZipFinder
from .packaging import CAPSULE_FILE_NAME, META_FILE_NAME

MAJOR_COMPATIBLE_VERSION = 0
MINOR_COMPATIBLE_VERSION = 1
"""The capsule version that this version of the library supports."""

MAJOR_MINOR_SEMVER_PATTERN = re.compile(r"([0-9]+)\.([0-9]+)")


def load_capsule(path: Union[str, Path],
                 key=None, inference_mode=True) -> BaseCapsule:
    """Load a capsule from the filesystem.

    :param path: The path to the capsule file
    :param key: The AES key to decrypt the capsule with, or None if the capsule
        is not encrypted
    :param inference_mode: If True, the backends for this capsule will be
        started. If False, the capsule will never be able to run inference, but
        it will still have it's various readable attributes.
    """
    path = Path(path)
    loaded_files = {}

    if key is None:
        # Capsule is unencrypted and already a zip file
        capsule_data = path.read_bytes()
    else:
        # Decrypt the capsule into its original form, a zip file
        capsule_data = decrypt_file(path, key)
    file_like = BytesIO(capsule_data)

    code = None
    with ZipFile(file_like, "r") as capsule_file:
        if CAPSULE_FILE_NAME not in capsule_file.namelist():
            raise RuntimeError(f"Capsule {path} has no {CAPSULE_FILE_NAME}")

        if META_FILE_NAME not in capsule_file.namelist():
            raise IncompatibleCapsuleError(
                f"Capsule {path} has no {META_FILE_NAME}")

        for name in capsule_file.namelist():
            if name == CAPSULE_FILE_NAME:
                # Every capsule has a capsule.py file defining the capsule's
                # behavior
                code = capsule_file.read(CAPSULE_FILE_NAME)
            else:
                # Load all other files as well
                loaded_files[name] = capsule_file.read(name)

        # Read the meta.conf and get the OpenVisionCapsules API compatibility
        # version
        meta_conf = configparser.ConfigParser()
        meta_conf.read_string(loaded_files[META_FILE_NAME].decode("utf-8"))
        compatibility_version = meta_conf["about"]["api_compatibility_version"]

        match = MAJOR_MINOR_SEMVER_PATTERN.fullmatch(compatibility_version)
        if match is None:
            raise InvalidCapsuleError(
                f"Invalid API compatibility version format "
                f"'{compatibility_version}'. Version must be in the format "
                f"'[major].[minor]'.")
        try:
            major, minor = map(int, (match[1], match[2]))
        except ValueError:
            raise InvalidCapsuleError(
                f"Compatibility versions must be numbers, got "
                f"{major}.{minor}.")
        if major != MAJOR_COMPATIBLE_VERSION:
            raise IncompatibleCapsuleError(
                f"Capsule {path} is not compatible with this software. The "
                f"capsule's OpenVisionCapsules required major version is "
                f"{major} but this software uses OpenVisionCapsules "
                f"{MAJOR_COMPATIBLE_VERSION}.{MINOR_COMPATIBLE_VERSION}.")
        if minor > MINOR_COMPATIBLE_VERSION:
            raise IncompatibleCapsuleError(
                f"Capsule {path} requires a version of OpenVisionCapsules "
                f"that is too new for this software. The capsule requires at "
                f"least version {major}.{minor} but this software uses "
                f"OpenVisionCapsules "
                f"{MAJOR_COMPATIBLE_VERSION}.{MINOR_COMPATIBLE_VERSION}.")

        # With the capsule's code loaded, initialize the object
        capsule_module = ModuleType(path.stem)
        try:
            # Allow the capsule.py to import other files in the capsule
            capsule_dir_path = (path.parent / path.stem).absolute()
            sys.meta_path.insert(1, ZipFinder(capsule_file, capsule_dir_path))

            # Run the capsule
            compiled = compile(code, capsule_dir_path / "capsule.py", "exec")
            exec(compiled, capsule_module.__dict__)
        except Exception as e:
            raise InvalidCapsuleError(
                "Could not execute the code in the capsule!\n"
                f"File: {path}\n"
                f"Error: {e}")
        finally:
            # Remove custom import code
            sys.meta_path.pop(1)

    # noinspection PyUnresolvedReferences
    new_capsule: BaseCapsule = capsule_module.Capsule(
        capsule_files=loaded_files,
        inference_mode=inference_mode)

    try:
        _validate_capsule(new_capsule)
    except InvalidCapsuleError as e:
        logging.warning(f"Failed to load capsule {path}")
        new_capsule.close()
        raise e

    return new_capsule


def _validate_capsule(capsule: BaseCapsule):
    """This will try calling different attributes on a capsule to make sure
    they are there. This is based on the API compatibility version.

    :raises: InvalidCapsuleError
    """

    def check_arg_names(func: Callable,
                        correct: List[str],
                        ignore: Optional[List[str]] = None) \
            -> bool:
        """Return False if a function has the wrong argument names. Return
        true if they are correct.
        Usage:
        >>> def my_func(self, frame, detection_node):
        ...     pass
        >>> check_arg_names(my_func, ['self'], ['frame', 'detection_node'])
        True
        """
        # noinspection PyUnresolvedReferences
        code = func.__code__

        ignore = [] if ignore is None else ignore
        all_var_names = code.co_varnames
        arg_names = all_var_names[:code.co_argcount]
        filtered = [n for n in arg_names if n not in ignore]
        return filtered == correct

    try:
        # If the ASCII flag is used, only [a-zA-Z0-9_] is matched for \w.
        # Otherwise, some other Unicode characters can be matched, depending on
        # the locale. We don't want that.
        if re.fullmatch(r"\w+", capsule.name, flags=re.ASCII) is None:
            raise InvalidCapsuleError(
                "Capsule names must only contain alphanumeric characters and "
                "underscores")

        # Validate the capsule class attributes
        capsule_assertions = [
            isinstance(capsule.name, str),
            callable(capsule.backend_loader),
            isinstance(capsule.version, int),
            isinstance(capsule.input_type, NodeDescription),
            isinstance(capsule.output_type, NodeDescription),
            isinstance(capsule.options, dict)]

        if not all(capsule_assertions):
            raise InvalidCapsuleError(
                f"The capsule has an invalid internal configuration!\n" +
                f"Capsule Assertions: {capsule_assertions}")

        # Make sure that certain things are NOT attributes (we don't want
        # accidental leftover code from previous capsule versions)
        unwanted_attributes = ["backend_config"]
        for unwanted_attr in unwanted_attributes:
            try:
                # This should throw an attribute error
                capsule.__getattribute__(unwanted_attr)
                raise InvalidCapsuleError(
                    f"The capsule has leftover attributes from a previous "
                    f"OpenVisionCapsules API version. Attribute name: "
                    f"{unwanted_attr}")
            except AttributeError:
                pass

        # Validate the capsule's backend_loader takes the right args
        loader = capsule.backend_loader
        loader_assertions = [
            callable(loader),
            check_arg_names(func=loader, correct=["capsule_files", "device"])]
        if not all(loader_assertions):
            raise InvalidCapsuleError(
                f"The capsule's backend_loader has an invalid configuration!\n"
                f"Loader Assertions: {loader_assertions}")

        # Validate the backend class attributes
        if capsule.backends is not None:
            backend = capsule.backends[0]
            backend_assertions = [
                callable(backend.batch_predict),
                callable(backend.process_frame),
                callable(backend.close),
                isinstance(capsule.backends[0], BaseBackend)]
            if not all(backend_assertions):
                raise InvalidCapsuleError(
                    f"The capsule's backend has an invalid configuration!\n"
                    f"Backend Assertions: {backend_assertions}")

        # Validate the stream state
        stream_state = capsule.stream_state
        stream_state_assertions = [
            (stream_state is BaseStreamState or
             BaseStreamState in stream_state.__bases__)]
        if not all(stream_state_assertions):
            raise InvalidCapsuleError(
                "The capsule's stream_state has an invalid configuration!\n"
                f"Stream State Assertions: {stream_state_assertions}")

        # Validate that if the capsule is an encoder, it has a threshold option
        if capsule.capability.encoded:
            if "recognition_threshold" not in capsule.options.keys():
                raise InvalidCapsuleError(
                    "This capsule can encode, but doesn't have a option "
                    "called \"recognition_threshold\"!")
    except InvalidCapsuleError:
        # Don't catch this exception type in the "except Exception" below
        raise
    except AttributeError as e:
        message = f"The capsule did not describe the necessary attributes. " \
                  f"Error: {e}"
        raise InvalidCapsuleError(message)
    except Exception as e:
        message = f"This capsule is invalid for unknown reasons. Error: {e}"
        raise InvalidCapsuleError(message)



class IncompatibleCapsuleError(Exception):
    """This is thrown when a capsule is of an incompatible API version for this
    version of the library.
    """


class InvalidCapsuleError(Exception):
    """This is thrown when a capsule has the correct API compatibility version,
    but it still doesn't have the correct information on it.
    """



"""Contains utility functions for doing data modification work in a capsule."""
import math
from enum import auto, Enum
from typing import List, Optional, Tuple, Union

import cv2
import numpy as np

from vcap.detection_node import BoundingBox, DetectionNode


def _clamp(num, min, max):
    if num < min:
        return 0
    if num > max:
        return max - 1
    return num


class Crop:
    """A crop that may be applied to a frame.

    Usage Example:
    >>> det_node = DetectionNode(...)
    >>> frame: np.ndarray
    >>> cropped = Crop.from_detection(det_node).pad_percent(top=10).apply(frame)
    """

    def __init__(self, x1, y1, x2, y2):
        self.x1 = x1
        self.y1 = y1
        self.x2 = x2
        self.y2 = y2

    def pad_percent(self, top=0, bottom=0, left=0, right=0) -> "Crop":
        width = self.x2 - self.x1
        height = self.y2 - self.y1

        self.y1 -= height * (top / 100)
        self.y2 += height * (bottom / 100)
        self.x1 -= width * (left / 100)
        self.x2 += width * (right / 100)

        return self

    def pad_px(self, top=0, bottom=0, left=0, right=0) -> "Crop":
        self.y1 -= top
        self.y2 += bottom
        self.x1 -= left
        self.x2 += right

        return self

    def apply(self, frame: np.ndarray) -> np.ndarray:
        self.y1 = int(round(_clamp(self.y1, 0, frame.shape[0])))
        self.y2 = int(round(_clamp(self.y2, 0, frame.shape[0])))
        self.x1 = int(round(_clamp(self.x1, 0, frame.shape[1])))
        self.x2 = int(round(_clamp(self.x2, 0, frame.shape[1])))
        return frame[self.y1:self.y2, self.x1:self.x2, :]

    @classmethod
    def from_detection(cls, node: DetectionNode) -> "Crop":
        """Uses the given detection area to create a crop."""
        bbox = node.bbox
        return cls(bbox.x1, bbox.y1, bbox.x2, bbox.y2)


class Clamp:
    """Scales down a frame if its size is over a given amount, preserving the
    original aspect ratio. It can then scale up any detections that have been
    run on the smaller frame to fit the original frame size.

    If the frame is under the given amount, nothing is done and the frame is
    kept the same.

    Usage Example:
    >>> clamp = Clamp(frame, 100, 100)
    >>> detection_nodes = some_detection_inference(clamp.apply())
    >>> clamp.scale_detection_nodes(detection_nodes)
    """

    def __init__(self, frame, max_width, max_height):
        """
        :param frame: The frame to scale
        :param max_width: The maximum allowed width of the frame
        :param max_height: The maximum allowed height of the frame
        """
        self.frame = frame
        self.max_width = max_width
        self.max_height = max_height

        height = float(self.frame.shape[0])
        width = float(self.frame.shape[1])

        if height <= self.max_height and width <= self.max_width:
            # Image is already small enough
            self.scale_factor = 1.0
        elif height >= width:
            self.scale_factor = self.max_height / height
        else:
            self.scale_factor = self.max_width / width

    def apply(self) -> np.ndarray:
        """
        :return: A clamped version of the frame
        """
        height = float(self.frame.shape[0])
        width = float(self.frame.shape[1])

        new_width = int(width * self.scale_factor)
        new_height = int(height * self.scale_factor)
        return cv2.resize(self.frame, (new_width, new_height))

    def scale_detection_nodes(self, nodes: List[DetectionNode]):
        """Scales up the given DetectionNode objects to fit the original frame
        size.
        """
        node_scale = 1.0 / self.scale_factor
        for node in nodes:
            node.scale(node_scale, node_scale)
        return nodes


class Resize:
    """Resize frames to a specified size

    Allows for maintaining aspect ratio
    Allows for padding
    Allows for cropping

    Usage example:
    >>> # Resize up, then crop right/bottom side to be 480p
    >>> frame: np.ndarray
    >>> frame = (Resize(frame) \
    ...          .resize(1920, 1080, Resize.ResizeType.FIT_BOTH) \
    ...          .crop(640, 480, Resize.CropPadType.RIGHT_BOTTOM) \
    ...          .frame)

    >>> # Resize down, then pad all around to be back to 1080p
    >>> frame: np.ndarray
    >>> frame = (Resize(frame)
    ...          .resize(640, 480, Resize.ResizeType.FIT_BOTH)
    ...          .pad(1920, 1080, Resize.CropPadType.BOTH)
    ...          .frame)
    """

    class ResizeType(Enum):
        WIDTH = auto()
        """Fit width exactly, maintain aspect ratio"""
        HEIGHT = auto()
        """Fit height exactly, maintain aspect ratio"""
        FIT_BOTH = auto()
        """Fit both height and width within bounds, but only one will be exact.
        Will end up smaller than (or same as) requested size.
        Basically the opposite of FIT_ONE in implementation.
        """
        FIT_ONE = auto()
        """Fit one of height or width within size, maintaining aspect ratio.
        Will end up bigger than (or same as) requested size.
        Basically the opposite of FIT_BOTH in implementation.
        Usually followed by a crop.
        """
        EXACT = auto()
        """Resize image exactly, ignoring aspect ratio. crop and pad attributes
        will have no effect"""
        NONE = auto()
        """Don't resize"""

    class CropPadType(Enum):
        RIGHT_BOTTOM = auto()
        """Crop/pad the larger index pixels"""
        LEFT_TOP = auto()
        """Crop/pad the smaller index pixels"""
        ALL = auto()
        """Crop/pad evenly from both sides. Smaller index side will be one
        greater than larger other if odd crop/pad)"""
        CROP_START_POINT = auto()
        """Crop to size with top left of result at top_left. Not available for
        padding"""
        NONE = auto()
        """Don't crop/pad"""

    class _OperationType(Enum):
        SCALE = auto()
        OFFSET = auto()

    def __init__(self, frame: np.ndarray):

        # Used to keep track of operations so that scale_and_offset_detection
        # nodes works right
        self._operations: List[Tuple[Resize._OperationType,
                                     Union[Tuple[int, int],
                                           Tuple[float, float]]]] = []

        self.frame = frame

    def resize(self, resize_width: int, resize_height: int,
               resize_type: ResizeType):

        frame_width = self.frame.shape[1]
        frame_height = self.frame.shape[0]

        new_width = frame_width
        new_height = frame_height

        if resize_type is self.ResizeType.WIDTH:
            # Resize height at same ratio as width to preserve aspect ratio
            new_width = resize_width
            new_height = (resize_width / frame_width) * frame_height

        elif resize_type is self.ResizeType.HEIGHT:
            # Resize width at same ratio as height to preserve aspect ratio
            new_height = resize_height
            new_width = (resize_height / frame_height) * frame_width

        elif resize_type is self.ResizeType.FIT_BOTH:

            if resize_width / resize_height > frame_width / frame_height:
                # Target wider than source. Will be limited by height
                new_height = resize_height
                new_width = (resize_height / frame_height) * frame_width
            else:
                # Target taller than source. Will be limited by width
                new_width = resize_width
                new_height = (resize_width / frame_width) * frame_height

        elif resize_type is self.ResizeType.FIT_ONE:

            if resize_width / resize_height > frame_width / frame_height:
                # Target wider than source. Will be limited by width
                new_width = resize_width
                new_height = (resize_width / frame_width) * frame_height
            else:
                # Target taller than source. Will be limited by height
                new_height = resize_height
                new_width = (resize_height / frame_height) * frame_width

        elif resize_type is self.ResizeType.EXACT:
            new_width = resize_width
            new_height = resize_height

        elif resize_type is self.ResizeType.NONE:
            pass

        else:
            raise ValueError("Invalid resize type")

        new_width = round(new_width)
        new_height = round(new_height)

        # Account for scaling
        scale_width = new_width / frame_width
        scale_height = new_height / frame_height
        self._operations.append(
            (self._OperationType.SCALE, (scale_width, scale_height))
        )

        self.frame = cv2.resize(self.frame, (new_width, new_height))

        return self

    def crop_bbox(self, bbox: BoundingBox):
        return self.crop(
            crop_width=int(bbox.width),
            crop_height=int(bbox.height),
            crop_type=self.CropPadType.CROP_START_POINT,
            top_left=(int(bbox.x1), int(bbox.y1)))

    def crop(self, crop_width: int, crop_height: int, crop_type: CropPadType,
             top_left: Optional[Tuple[int, int]] = None):
        """Crop the frame

        :param crop_width: Resulting width of the frame after cropping
        :param crop_height: Resulting height of the frame after cropping
        :param crop_type: Type of crop to perform
        :param top_left: [Optional] Point in input's coordinate space that will
        serve as the top left point in the output. Only used if
        CropPadType=CROP_START_POINT. Tuple in form (x, y)
        """

        frame_width = self.frame.shape[1]
        frame_height = self.frame.shape[0]

        start_x = 0
        end_x = frame_width
        start_y = 0
        end_y = frame_height

        if crop_type is self.CropPadType.LEFT_TOP:
            start_x = frame_width - crop_width
            start_y = frame_height - crop_height

        elif crop_type is self.CropPadType.RIGHT_BOTTOM:
            end_x = crop_width
            end_y = crop_height

        elif crop_type is self.CropPadType.ALL:
            start_x = (frame_width - crop_width) // 2
            start_y = (frame_height - crop_height) // 2

            end_x = start_x + crop_width
            end_y = start_y + crop_height

        elif crop_type is self.CropPadType.CROP_START_POINT:
            start_x, start_y = top_left
            end_x = start_x + crop_width
            end_y = start_y + crop_height

        elif crop_type is self.CropPadType.NONE:
            pass

        else:
            raise ValueError("Invalid crop type")

        start_x = max(0, start_x)
        start_y = max(0, start_y)
        end_x = min(frame_width, end_x)  # Not really necessary for splicing
        end_y = min(frame_height, end_y)

        # Account for cropping displacements
        self._operations.append(
            (self._OperationType.OFFSET, (start_x, start_y))
        )

        self.frame = self.frame[start_y:end_y, start_x:end_x]

        return self

    def pad(self, pad_width, pad_height, pad_value,
            pad_type: CropPadType):
        left_pad = 0
        right_pad = 0
        top_pad = 0
        bottom_pad = 0

        frame_width = self.frame.shape[1]
        frame_height = self.frame.shape[0]

        if pad_type is self.CropPadType.LEFT_TOP:
            left_pad = pad_width - frame_width
            top_pad = pad_height - frame_height

        elif pad_type is self.CropPadType.RIGHT_BOTTOM:
            right_pad = pad_width - frame_width
            bottom_pad = pad_height - frame_height

        elif pad_type is self.CropPadType.ALL:
            left_pad = (pad_width - frame_width) // 2
            right_pad = (pad_width - frame_width) - left_pad

            top_pad = (pad_height - frame_height) // 2
            bottom_pad = (pad_height - frame_height) - top_pad

        elif pad_type is self.CropPadType.NONE:
            pass

        else:
            raise ValueError("Invalid pad type")

        left_pad = max(0, left_pad)
        right_pad = max(0, right_pad)
        top_pad = max(0, top_pad)
        bottom_pad = max(0, bottom_pad)

        # Account for padding displacements
        self._operations.append(
            (self._OperationType.OFFSET, (-left_pad, -top_pad))
        )

        self.frame = cv2.copyMakeBorder(
            self.frame,
            top_pad, bottom_pad, left_pad, right_pad,
            cv2.BORDER_CONSTANT,
            value=pad_value)

        return self

    def scale_and_offset_detection_nodes(self, nodes: List[DetectionNode]):
        """Scales up the given DetectionNode objects to fit the original frame
        size.
        """
        for node in nodes:
            for operation, (x, y) in self._operations[::-1]:
                if operation is self._OperationType.SCALE:
                    node.scale(1 / x, 1 / y)
                elif operation is self._OperationType.OFFSET:
                    node.offset(-x, -y)
        return nodes


class SizeFilter:
    """Filters detection nodes based on if they're large or small enough.

    Usage example:
    >>> nodes: List[DetectionNode]
    >>> filtered = (SizeFilter(nodes)
    ...             .min_size(10, 10)
    ...             .max_size(1000, 1000)
    ...             .apply())
    """

    def __init__(self, nodes: List[DetectionNode]):
        self.nodes = nodes

        self._min_width = 0
        self._min_height = 0
        self._min_area = 0

        self._max_width = math.inf
        self._max_height = math.inf
        self._max_area = math.inf

    def min_size(self, width, height):
        self._min_width = width
        self._min_height = height
        return self

    def max_size(self, width, height):
        self._max_width = width
        self._max_height = height
        return self

    def min_area(self, area):
        self._min_area = area
        return self

    def max_area(self, area):
        self._max_area = area
        return self

    def apply(self) -> List[DetectionNode]:
        return [node for node in self.nodes if self._within_size(node)]

    def _within_size(self, node: DetectionNode):
        size = node.bbox.size
        return self._min_width <= size[0] <= self._max_width \
               and self._min_height <= size[1] <= self._max_height \
               and self._min_area <= size[0] * size[1] <= self._max_area



import abc
from queue import Queue
from typing import Any, Dict, List

import numpy as np

from vcap.node_description import DETECTION_NODE_TYPE
from vcap.ovens import Oven
from vcap.options import OPTION_TYPE
from vcap.stream_state import BaseStreamState


class BaseBackend(abc.ABC):
    """An object that provides low-level prediction functionality for batches
    of frames.
    """

    def __init__(self):
        self.oven = Oven(self.batch_predict)

    def send_to_batch(self, input_data: Any) -> Queue:
        """Sends the given object to the batch_predict method for processing.
        This call does not block. Instead, the result will be provided on the
        returned queue. The batch_predict method must be overridden on the
        backend this method is being called on.

        :param input_data: The input object to send to batch_predict
        :return: A queue where results will be stored
        """
        return self.oven.submit(input_data)

    @abc.abstractmethod
    def process_frame(self,
                      frame: np.ndarray,
                      detection_node: DETECTION_NODE_TYPE,
                      options: Dict[str, OPTION_TYPE],
                      state: BaseStreamState) \
            -> DETECTION_NODE_TYPE:
        """A method that does the preprocessing, inference, and postprocessing
        work for a frame. It has the ability to call self.send_to_batch(frame)
        to send work for batching (eventually run in batch_predict)."""

    def batch_predict(self, input_data_list: List[Any]) -> List[Any]:
        """Runs prediction on a batch of frames (or objects). This method must
        be overridden for capsules that use send_to_batch.

        :param input_data_list: A list of objects. Whatever the model requires
                                for each frame.
        """
        raise NotImplementedError(
            "Attempt to do batch prediction on a Backend that does not have "
            "the batch_predict method defined. Did you call send_to_batch on "
            "a backend that does not override batch_predict?")

    @abc.abstractmethod
    def close(self) -> None:
        """De-initializes the backend."""
        pass



from io import BytesIO
from pathlib import Path
from zipfile import ZipFile

from .crypto_utils import encrypt

CAPSULE_EXTENSION = ".cap"
"""The file extension that capsule files use."""

CAPSULE_FILE_NAME = "capsule.py"
"""The name of the main file in the capsule."""

META_FILE_NAME = "meta.conf"
"""The name of the meta file that contains information such as the API 
compatibility version of this capsule"""


class UnpackagedCapsuleError(Exception):
    pass


def package_capsule(unpackaged_dir: Path, output_file: Path, key=None):
    """Packages an unpackaged capsule up into a zip and applies encryption.

    :param unpackaged_dir: The directory to package up
    :param output_file: The name and location to save the packaged capsule
    :param key: An AES key to encrypt the capsule with
    """
    # Check to make sure there is a capsule.py
    if len(list(unpackaged_dir.glob(CAPSULE_FILE_NAME))) == 0:
        raise UnpackagedCapsuleError(
            f"Unpackaged capsule {unpackaged_dir} is missing a "
            f"{CAPSULE_FILE_NAME}")
    # Check to make sure there's a meta.conf
    if len(list(unpackaged_dir.glob(META_FILE_NAME))) == 0:
        raise UnpackagedCapsuleError(
            f"Unpackaged capsule {unpackaged_dir} is missing a "
            f"{META_FILE_NAME}")

    saved_zip_bytes = BytesIO()
    with ZipFile(saved_zip_bytes, "w") as capsule_file:
        for f in unpackaged_dir.glob("**/*"):
            relative_name = f.relative_to(unpackaged_dir)
            capsule_file.write(f, relative_name)

    with output_file.open("wb") as output:
        # Move the read cursor to the start of the zip data
        saved_zip_bytes.seek(0)
        if key is None:
            output.write(saved_zip_bytes.read())
        else:
            encrypt(key, saved_zip_bytes, output)
